@@ -102,9 +102,6 @@ static fsal_status_t lookup(struct fsal_obj_handle *dir_pub,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&st, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*obj_pub = &obj->handle;
@@ -250,8 +247,14 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 	struct Inode *i = NULL;
 	fsal_status_t status;
 
+	LogFullDebug(COMPONENT_FSAL,
+		     "mode = %o uid=%d gid=%d",
+		     attrib->mode, (int) op_ctx->creds->caller_uid,
+		     (int) op_ctx->creds->caller_gid);
+
 	unix_mode = fsal2unix_mode(attrib->mode)
 		& ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
+
 	rc = ceph_ll_mkdir(export->cmount, dir->i, name, unix_mode, &st, &i,
 			   op_ctx->creds->caller_uid,
 			   op_ctx->creds->caller_gid);
@@ -264,9 +267,9 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 	*new_obj = &obj->handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -289,13 +292,10 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes(&st, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
 		}
 	}
 
-	FSAL_SET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
 	return status;
 }
@@ -381,10 +381,10 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 
 	*new_obj = &obj->handle;
 
-	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	/* We handled the mode and rawdev above. */
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE | ATTR_RAWDEV);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -407,13 +407,10 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes(&st, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
 		}
 	}
 
-	FSAL_SET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
 	return status;
 #else
@@ -472,9 +469,9 @@ static fsal_status_t ceph_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 	*new_obj = &obj->handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -497,13 +494,10 @@ static fsal_status_t ceph_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes(&st, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
 		}
 	}
 
-	FSAL_SET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
 	return status;
 }
@@ -576,18 +570,15 @@ static fsal_status_t getattrs(struct fsal_obj_handle *handle_pub,
 	rc = ceph_ll_getattr(export->cmount, handle->i, &st, 0, 0);
 
 	if (rc < 0) {
-		if (attrs->mask & ATTR_RDATTR_ERR) {
+		if (attrs->request_mask & ATTR_RDATTR_ERR) {
 			/* Caller asked for error to be visible. */
-			attrs->mask = ATTR_RDATTR_ERR;
+			attrs->valid_mask = ATTR_RDATTR_ERR;
 		}
 		return ceph2fsal_error(rc);
 	}
 
 	posix2fsal_attributes(&st, attrs);
 
-	/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-	attrs->mask &= ~ATTR_RDATTR_ERR;
-
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
@@ -1039,7 +1030,7 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 			/* We need to use the global fd to continue, and take
 			 * the lock to protect it.
 			 */
-			my_fd = &hdl->fd;
+			my_fd = &myself->fd;
 			PTHREAD_RWLOCK_wrlock(&obj_hdl->lock);
 		}
 
@@ -1089,14 +1080,19 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 
 		if (state == NULL) {
 			/* If no state, release the lock taken above and return
-			 * status.
+			 * status. If success, we haven't done any permission
+			 * check so ask the caller to do so.
 			 */
 			PTHREAD_RWLOCK_unlock(&obj_hdl->lock);
+			*caller_perm_check = !FSAL_IS_ERROR(status);
 			return status;
 		}
 
 		if (!FSAL_IS_ERROR(status)) {
-			/* Return success. */
+			/* Return success. We haven't done any permission
+			 * check so ask the caller to do so.
+			 */
+			*caller_perm_check = true;
 			return status;
 		}
 
@@ -1156,9 +1152,6 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 			LogFullDebug(COMPONENT_FSAL,
 				     "open returned %s",
 				     fsal_err_txt(status));
-		} else {
-			/* No permission check was actually done... */
-			*caller_perm_check = true;
 		}
 
 		return status;
@@ -1181,10 +1174,10 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 		    ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
 		/* Don't set the mode if we later set the attributes */
-		FSAL_UNSET_MASK(attrib_set->mask, ATTR_MODE);
+		FSAL_UNSET_MASK(attrib_set->valid_mask, ATTR_MODE);
 	}
 
-	if (createmode == FSAL_UNCHECKED && (attrib_set->mask != 0)) {
+	if (createmode == FSAL_UNCHECKED && (attrib_set->valid_mask != 0)) {
 		/* If we have FSAL_UNCHECKED and want to set more attributes
 		 * than the mode, we attempt an O_EXCL create first, if that
 		 * succeeds, then we will be allowed to set the additional
@@ -1219,7 +1212,7 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 		posix_flags &= ~O_EXCL;
 		retval =
 		    ceph_ll_create(export->cmount,  myself->i, name, unix_mode,
-				   posix_flags, &stat, &i, &my_fd->fd,
+				   posix_flags, &stat, &i, &fd,
 				   op_ctx->creds->caller_uid,
 				   op_ctx->creds->caller_gid);
 		if (retval < 0) {
@@ -1237,12 +1230,22 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 	 * Note that in an UNCHECKED retry we MIGHT have re-created the
 	 * file and won't remember that. Oh well, so in that rare case we
 	 * leak a partially created file if we have a subsequent error in here.
-	 * Since we were able to do the permission check even if we were not
-	 * creating the file, let the caller know the permission check has
-	 * already been done. Note it IS possible in the case of a race between
-	 * an UNCHECKED open and an external unlink, we did create the file.
 	 */
 	created = (posix_flags & O_EXCL) != 0;
+
+	/** @todo FSF: Note that the current implementation of ceph_ll_create
+	 *             does not accept an alt groups list, so it is possible
+	 *             a create (including an UNCHECKED create on an already
+	 *             existing file) would fail because the directory or
+	 *             file was owned by a group other than the primary group.
+	 *             Conversely, it could also succeed when it should have
+	 *             failed if other is granted more permission than
+	 *             one of the alt groups).
+	 */
+
+	/* Since we did the ceph_ll_create using the caller's credentials,
+	 * we don't need to do an additional permission check.
+	 */
 	*caller_perm_check = false;
 
 	construct_handle(&stat, i, export, &hdl);
@@ -1261,7 +1264,7 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 
 	*new_obj = &hdl->handle;
 
-	if (created && attrib_set->mask != 0) {
+	if (created && attrib_set->valid_mask != 0) {
 		/* Set attributes using our newly opened file descriptor as the
 		 * share_fd if there are any left to set (mode and truncate
 		 * have already been handled).
@@ -1285,7 +1288,7 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 			status = (*new_obj)->obj_ops.getattrs(*new_obj,
 							      attrs_out);
 			if (FSAL_IS_ERROR(status) &&
-			    (attrs_out->mask & ATTR_RDATTR_ERR) == 0) {
+			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
 				 * to get the attributes. Otherwise continue
 				 * with attrs_out indicating ATTR_RDATTR_ERR.
@@ -1299,9 +1302,6 @@ fsal_status_t ceph_open2(struct fsal_obj_handle *obj_hdl,
 		 * we used to create the fsal_obj_handle.
 		 */
 		posix2fsal_attributes(&stat, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	if (state != NULL) {
@@ -1886,7 +1886,7 @@ fsal_status_t ceph_lock_op2(struct fsal_obj_handle *obj_hdl,
  * @brief Set attributes on an object
  *
  * This function sets attributes on an object.  Which attributes are
- * set is determined by attrib_set->mask. The FSAL must manage bypass
+ * set is determined by attrib_set->valid_mask. The FSAL must manage bypass
  * or not of share reservations, and a state may be passed.
  *
  * @param[in] obj_hdl    File on which to operate
@@ -1913,11 +1913,11 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 	/* Mask of attributes to set */
 	uint32_t mask = 0;
 
-	if (attrib_set->mask & ~settable_attributes) {
+	if (attrib_set->valid_mask & ~settable_attributes) {
 		LogDebug(COMPONENT_FSAL,
 			 "bad mask %"PRIx64" not settable %"PRIx64,
-			 attrib_set->mask,
-			 attrib_set->mask & ~settable_attributes);
+			 attrib_set->valid_mask,
+			 attrib_set->valid_mask & ~settable_attributes);
 		return fsalstat(ERR_FSAL_INVAL, 0);
 	}
 
@@ -1925,14 +1925,14 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 		    "attrs ", attrib_set, false);
 
 	/* apply umask, if mode attribute is to be changed */
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MODE))
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MODE))
 		attrib_set->mode &=
 		    ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
 	/* Test if size is being set, make sure file is regular and if so,
 	 * require a read/write file descriptor.
 	 */
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_SIZE)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_SIZE)) {
 		if (obj_hdl->type != REGULAR_FILE) {
 			LogFullDebug(COMPONENT_FSAL,
 				     "Setting size on non-regular file");
@@ -1956,7 +1956,7 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 
 	memset(&st, 0, sizeof(struct stat));
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_SIZE)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_SIZE)) {
 		rc = ceph_ll_truncate(export->cmount, myself->i,
 				      attrib_set->filesize, 0, 0);
 
@@ -1969,27 +1969,27 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 		}
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MODE)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MODE)) {
 		mask |= CEPH_SETATTR_MODE;
 		st.st_mode = fsal2unix_mode(attrib_set->mode);
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_OWNER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_OWNER)) {
 		mask |= CEPH_SETATTR_UID;
 		st.st_uid = attrib_set->owner;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_GROUP)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_GROUP)) {
 		mask |= CEPH_SETATTR_GID;
 		st.st_gid = attrib_set->group;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ATIME)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ATIME)) {
 		mask |= CEPH_SETATTR_ATIME;
 		st.st_atim = attrib_set->atime;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ATIME_SERVER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ATIME_SERVER)) {
 		mask |= CEPH_SETATTR_ATIME;
 		struct timespec timestamp;
 
@@ -2004,11 +2004,11 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 		st.st_atim = timestamp;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MTIME)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MTIME)) {
 		mask |= CEPH_SETATTR_MTIME;
 		st.st_mtim = attrib_set->mtime;
 	}
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MTIME_SERVER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MTIME_SERVER)) {
 		mask |= CEPH_SETATTR_MTIME;
 		struct timespec timestamp;
 
@@ -2023,7 +2023,7 @@ fsal_status_t ceph_setattr2(struct fsal_obj_handle *obj_hdl,
 		st.st_mtim = timestamp;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_CTIME)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_CTIME)) {
 		mask |= CEPH_SETATTR_CTIME;
 		st.st_ctim = attrib_set->ctime;
 	}
