@@ -74,6 +74,7 @@ static void handle_release(struct fsal_obj_handle *obj_hdl)
 				"glfs_h_close returned error %s(%d)",
 				strerror(errno), errno);
 		}
+		objhandle->glhandle = NULL;
 	}
 
 	gsh_free(objhandle);
@@ -133,9 +134,6 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*handle = &objhandle->handle;
@@ -301,9 +299,6 @@ static fsal_status_t create(struct fsal_obj_handle *dir_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*handle = &objhandle->handle;
@@ -389,17 +384,14 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*handle = &objhandle->handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -411,6 +403,8 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 				     "setattr2 status=%s",
 				     fsal_err_txt(status));
 			(*handle)->obj_ops.release(*handle);
+			/* We released handle at this point */
+			glhandle = NULL;
 			*handle = NULL;
 		}
 	} else {
@@ -418,7 +412,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 		status.minor = 0;
 	}
 
-	FSAL_SET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
  out:
 	if (status.major != ERR_FSAL_NO_ERROR)
@@ -529,17 +523,14 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*handle = &objhandle->handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -551,6 +542,8 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 				     "setattr2 status=%s",
 				     fsal_err_txt(status));
 			(*handle)->obj_ops.release(*handle);
+			/* We released handle at this point */
+			glhandle = NULL;
 			*handle = NULL;
 		}
 	} else {
@@ -558,7 +551,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 		status.minor = 0;
 	}
 
-	FSAL_SET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
  out:
 	if (status.major != ERR_FSAL_NO_ERROR)
@@ -640,14 +633,11 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 	*handle = &objhandle->handle;
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -659,6 +649,8 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 				     "setattr2 status=%s",
 				     fsal_err_txt(status));
 			(*handle)->obj_ops.release(*handle);
+			/* We released handle at this point */
+			glhandle = NULL;
 			*handle = NULL;
 		}
 	} else {
@@ -765,6 +757,9 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 	 * way is found - that may involve introducing locks in
 	 * the gfapi's for close and getattrs etc.
 	 */
+
+	/** @todo: With support_ex() above may no longer be valid.
+	 * This needs to be revisited */
 	rc = glfs_h_stat(glfs_export->gl_fs,
 			 objhandle->glhandle, &buffxstat.buffstat);
 	if (rc != 0) {
@@ -773,9 +768,9 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 		else
 			status = gluster2fsal_error(errno);
 
-		if (attrs->mask & ATTR_RDATTR_ERR) {
+		if (attrs->request_mask & ATTR_RDATTR_ERR) {
 			/* Caller asked for error to be visible. */
-			attrs->mask = ATTR_RDATTR_ERR;
+			attrs->valid_mask = ATTR_RDATTR_ERR;
 		}
 		goto out;
 	}
@@ -786,8 +781,15 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 	else
 		buffxstat.is_dir = false;
 
-	status = glusterfs_get_acl(glfs_export, objhandle->glhandle,
-				   &buffxstat, attrs);
+	if (attrs->request_mask & ATTR_ACL) {
+		/* Fetch the ACL */
+		status = glusterfs_get_acl(glfs_export, objhandle->glhandle,
+					   &buffxstat, attrs);
+		if (!FSAL_IS_ERROR(status)) {
+			/* Success, so mark ACL as valid. */
+			attrs->valid_mask |= ATTR_ACL;
+		}
+	}
 
 	/* *
 	* The error ENOENT is not an expected error for GETATTRS
@@ -803,13 +805,10 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 	}
 
 	if (FSAL_IS_ERROR(status)) {
-		if (attrs->mask & ATTR_RDATTR_ERR) {
+		if (attrs->request_mask & ATTR_RDATTR_ERR) {
 			/* Caller asked for error to be visible. */
-			attrs->mask = ATTR_RDATTR_ERR;
+			attrs->valid_mask = ATTR_RDATTR_ERR;
 		}
-	} else {
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs->mask &= ~ATTR_RDATTR_ERR;
 	}
 
  out:
@@ -821,142 +820,6 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 	return status;
 }
 
-/**
- * @brief Implements GLUSTER FSAL objectoperation setattrs
- */
-
-static fsal_status_t setattrs(struct fsal_obj_handle *obj_hdl,
-			      struct attrlist *attrs)
-{
-	int rc = 0;
-	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
-	glusterfs_fsal_xstat_t buffxstat;
-	int mask = 0;
-	int attr_valid = 0;
-	struct glusterfs_export *glfs_export =
-	    container_of(op_ctx->fsal_export, struct glusterfs_export, export);
-	struct glusterfs_handle *objhandle =
-	    container_of(obj_hdl, struct glusterfs_handle, handle);
-#ifdef GLTIMING
-	struct timespec s_time, e_time;
-
-	now(&s_time);
-#endif
-	memset(&buffxstat, 0, sizeof(glusterfs_fsal_xstat_t));
-
-	/* sanity checks.
-	 * note : object_attributes is optional.
-	 */
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_SIZE)) {
-		rc = glfs_h_truncate(glfs_export->gl_fs, objhandle->glhandle,
-				     attrs->filesize);
-		if (rc != 0) {
-			status = gluster2fsal_error(errno);
-			goto out;
-		}
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_MODE)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MODE);
-		buffxstat.buffstat.st_mode = fsal2unix_mode(attrs->mode);
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_OWNER)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_UID);
-		buffxstat.buffstat.st_uid = attrs->owner;
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_GROUP)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_GID);
-		buffxstat.buffstat.st_gid = attrs->group;
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_ATIME)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_ATIME);
-		buffxstat.buffstat.st_atim = attrs->atime;
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_ATIME_SERVER)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_ATIME);
-		struct timespec timestamp;
-
-		rc = clock_gettime(CLOCK_REALTIME, &timestamp);
-		if (rc != 0) {
-			status = gluster2fsal_error(errno);
-			goto out;
-		}
-		buffxstat.buffstat.st_atim = timestamp;
-	}
-
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_MTIME)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MTIME);
-		buffxstat.buffstat.st_mtim = attrs->mtime;
-	}
-	if (FSAL_TEST_MASK(attrs->mask, ATTR_MTIME_SERVER)) {
-		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MTIME);
-		struct timespec timestamp;
-
-		rc = clock_gettime(CLOCK_REALTIME, &timestamp);
-		if (rc != 0) {
-			status = gluster2fsal_error(rc);
-			goto out;
-		}
-		buffxstat.buffstat.st_mtim = timestamp;
-	}
-
-	/** @todo: Check for attributes not supported and return */
-	/* EATTRNOTSUPP error.  */
-
-	if (NFSv4_ACL_SUPPORT) {
-		if (FSAL_TEST_MASK(attrs->mask, ATTR_ACL)) {
-			if (obj_hdl->type == DIRECTORY)
-				buffxstat.is_dir = true;
-			else
-				buffxstat.is_dir = false;
-
-			FSAL_SET_MASK(attr_valid, XATTR_ACL);
-			status =
-			  glusterfs_process_acl(glfs_export->gl_fs,
-						objhandle->glhandle,
-						attrs, &buffxstat);
-
-			if (FSAL_IS_ERROR(status))
-				goto out;
-			/* setting the ACL will set the */
-			/* mode-bits too if not already passed */
-			FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MODE);
-		}
-	} else if (FSAL_TEST_MASK(attrs->mask, ATTR_ACL)) {
-		status = fsalstat(ERR_FSAL_ATTRNOTSUPP, 0);
-		goto out;
-	}
-
-	/* If any stat changed, indicate that */
-	if (mask != 0)
-		FSAL_SET_MASK(attr_valid, XATTR_STAT);
-	if (FSAL_TEST_MASK(attr_valid, XATTR_STAT)) {
-		/*Only if there is any change in attrs send them down to fs */
-		rc = glfs_h_setattrs(glfs_export->gl_fs,
-				     objhandle->glhandle,
-				     &buffxstat.buffstat,
-				     mask);
-		if (rc != 0) {
-			status = gluster2fsal_error(errno);
-			goto out;
-		}
-	}
-
-	if (FSAL_TEST_MASK(attr_valid, XATTR_ACL))
-		status = glusterfs_set_acl(glfs_export, objhandle, &buffxstat);
-
-out:
-#ifdef GLTIMING
-	now(&e_time);
-	latency_update(&s_time, &e_time, lat_setattrs);
-#endif
-	return status;
-}
-
 /**
  * @brief Implements GLUSTER FSAL objectoperation link
  */
@@ -1286,6 +1149,12 @@ fsal_status_t glusterfs_close_func(struct fsal_obj_handle *obj_hdl,
 	return glusterfs_close_my_fd((struct glusterfs_fd *)fd);
 }
 
+/**
+ * @brief Find a file descriptor for a read, write, setattr2 operation.
+ *
+ * We do not need file descriptors for non-regular files, so this never has to
+ * handle them.
+ */
 fsal_status_t find_fd(struct glusterfs_fd *my_fd,
 		      struct fsal_obj_handle *obj_hdl,
 		      bool bypass,
@@ -1298,168 +1167,67 @@ fsal_status_t find_fd(struct glusterfs_fd *my_fd,
 {
 	struct glusterfs_handle *myself;
 	fsal_status_t status = {ERR_FSAL_NO_ERROR, 0};
-	int posix_flags;
 	struct glusterfs_fd  tmp_fd = {0}, *tmp2_fd = &tmp_fd;
-	struct glusterfs_export *glfs_export =
-	    container_of(op_ctx->fsal_export, struct glusterfs_export, export);
 
 	myself = container_of(obj_hdl, struct glusterfs_handle, handle);
 
-	fsal2posix_openflags(openflags, &posix_flags);
-
-	/* Handle non-regular files */
-	switch (obj_hdl->type) {
-	case SOCKET_FILE:
-	case CHARACTER_FILE:
-	case BLOCK_FILE:
-		/** @todo: check for O_NOACCESS. Refer vfs find_fd */
-		posix_flags = O_PATH;
-		break;
-
-	case REGULAR_FILE:
-		status = fsal_find_fd((struct fsal_fd **)&tmp2_fd, obj_hdl,
-				      (struct fsal_fd *)&myself->globalfd,
-				      &myself->share, bypass, state,
-				      openflags, glusterfs_open_func,
-				      glusterfs_close_func,
-				      has_lock, need_fsync,
-				      closefd, open_for_locks);
-
-		my_fd->glfd = tmp2_fd->glfd;
-		my_fd->openflags = tmp2_fd->openflags;
-		return status;
-
-	case SYMBOLIC_LINK:
-		posix_flags |= (O_PATH | O_RDWR | O_NOFOLLOW);
-		break;
-
-	case FIFO_FILE:
-		posix_flags |= O_NONBLOCK;
-		break;
-
-	case DIRECTORY:
-		my_fd->glfd = glfs_h_opendir(glfs_export->gl_fs,
-						myself->glhandle);
-		if (my_fd->glfd == NULL)
-			return gluster2fsal_error(errno);
-		*closefd = true;
-		return status;
-
-	case NO_FILE_TYPE:
-	case EXTENDED_ATTR:
+	/* Handle only regular files */
+	if (obj_hdl->type != REGULAR_FILE)
 		return fsalstat(posix2fsal_error(EINVAL), EINVAL);
-	}
 
-	/* Non-regular files */
-	status = glusterfs_open_my_fd(myself,
-				      openflags,
-				      posix_flags,
-				      &tmp_fd);
-	if (FSAL_IS_ERROR(status)) {
-		LogCrit(COMPONENT_FSAL,
-			 "Failed with %s openflags 0x%08x",
-			 strerror(errno), openflags);
-			return fsalstat(posix2fsal_error(errno), errno);
-	}
-
-	my_fd->glfd = tmp_fd.glfd;
-	my_fd->openflags = tmp_fd.openflags;
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "Opened glfd=%p for file of type %s",
-		     my_fd->glfd,
-		     object_file_type_to_str(obj_hdl->type));
-	*closefd = true;
+	status = fsal_find_fd((struct fsal_fd **)&tmp2_fd, obj_hdl,
+			      (struct fsal_fd *)&myself->globalfd,
+			      &myself->share, bypass, state,
+			      openflags, glusterfs_open_func,
+			      glusterfs_close_func,
+			      has_lock, need_fsync,
+			      closefd, open_for_locks);
 
+	my_fd->glfd = tmp2_fd->glfd;
+	my_fd->openflags = tmp2_fd->openflags;
 	return status;
 }
 
+/**
+ * @brief Merge a duplicate handle with an original handle
+ *
+ * This function is used if an upper layer detects that a duplicate
+ * object handle has been created. It allows the FSAL to merge anything
+ * from the duplicate back into the original.
+ *
+ * The caller must release the object (the caller may have to close
+ * files if the merge is unsuccessful).
+ *
+ * @param[in]  orig_hdl  Original handle
+ * @param[in]  dupe_hdl Handle to merge into original
+ *
+ * @return FSAL status.
+ *
+ */
 
-fsal_status_t glusterfs_fetch_attrs(struct glusterfs_handle *myself,
-				    struct glusterfs_fd *my_fd)
+fsal_status_t glusterfs_merge(struct fsal_obj_handle *orig_hdl,
+			      struct fsal_obj_handle *dupe_hdl)
 {
-	int retval = 0;
-	fsal_status_t status = {0, 0};
-	const char *func = "unknown";
-	glusterfs_fsal_xstat_t buffxstat;
-	struct attrlist *fsalattr;
-	struct glusterfs_export *glfs_export =
-	    container_of(op_ctx->fsal_export, struct glusterfs_export, export);
+	fsal_status_t status = {ERR_FSAL_NO_ERROR, 0};
 
+	if (orig_hdl->type == REGULAR_FILE &&
+	    dupe_hdl->type == REGULAR_FILE) {
+		/* We need to merge the share reservations on this file.
+		 * This could result in ERR_FSAL_SHARE_DENIED.
+		 */
+		struct glusterfs_handle *orig, *dupe;
 
-	/* Now stat the file as appropriate */
-	switch (myself->handle.type) {
-	case SOCKET_FILE:
-	case CHARACTER_FILE:
-	case BLOCK_FILE:
-	case SYMBOLIC_LINK: /** @todo: do we need glfd? */
-	case FIFO_FILE:
-		retval = glfs_h_stat(glfs_export->gl_fs, myself->glhandle,
-					&buffxstat.buffstat);
-		func = "stat";
-		break;
-	case REGULAR_FILE:
-	case DIRECTORY:
-		retval = glfs_fstat(my_fd->glfd, &buffxstat.buffstat);
-		func = "fstat";
-		break;
+		orig = container_of(orig_hdl, struct glusterfs_handle, handle);
+		dupe = container_of(dupe_hdl, struct glusterfs_handle, handle);
 
-	case NO_FILE_TYPE:
-	case EXTENDED_ATTR:
-		/* Caught during open with EINVAL */
-		break;
+		/* This can block over an I/O operation. */
+		PTHREAD_RWLOCK_wrlock(&orig_hdl->lock);
+
+		status = merge_share(&orig->share, &dupe->share);
+
+		PTHREAD_RWLOCK_unlock(&orig_hdl->lock);
 	}
 
-	if (retval < 0) {
-		if (errno == ENOENT)
-			retval = ESTALE;
-		else
-			retval = errno;
-
-		LogDebug(COMPONENT_FSAL, "%s failed with %s", func,
-			 strerror(retval));
-
-		if (myself->attributes.mask & ATTR_RDATTR_ERR) {
-			/* Caller asked for error to be visible. */
-			myself->attributes.mask = ATTR_RDATTR_ERR;
-		}
-
-		status = gluster2fsal_error(retval);
-		goto out;
-	}
-
-	fsalattr = &myself->attributes;
-	stat2fsal_attributes(&buffxstat.buffstat, &myself->attributes);
-
-	if (myself->handle.type == DIRECTORY)
-		buffxstat.is_dir = true;
-	else
-		buffxstat.is_dir = false;
-
-	status = glusterfs_get_acl(glfs_export, myself->glhandle,
-				   &buffxstat, fsalattr);
-
-	/*
-	 * The error ENOENT is not an expected error for GETATTRS
-	 * Due to this, operations such as RENAME will fail when
-	 * it calls GETATTRS on removed file. But for dead links
-	 * we should not return error
-	 */
-	if (status.minor == ENOENT) {
-		if (myself->handle.type == SYMBOLIC_LINK)
-			status = fsalstat(ERR_FSAL_NO_ERROR, 0);
-		else
-			status = gluster2fsal_error(ESTALE);
-	}
-
-out:
-	if (FSAL_IS_ERROR(status)
-		&& (myself->attributes.mask & ATTR_RDATTR_ERR)) {
-		myself->attributes.mask = ATTR_RDATTR_ERR;
-	} else {
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		myself->attributes.mask &= ~ATTR_RDATTR_ERR;
-	}
 	return status;
 }
 
@@ -1620,14 +1388,19 @@ static fsal_status_t glusterfs_open2(struct fsal_obj_handle *obj_hdl,
 
 		if (state == NULL) {
 			/* If no state, release the lock taken above and return
-			 * status.
+			 * status. If success, we haven't done any permission
+			 * check so ask the caller to do so.
 			 */
 			PTHREAD_RWLOCK_unlock(&obj_hdl->lock);
+			*caller_perm_check = !FSAL_IS_ERROR(status);
 			return status;
 		}
 
 		if (!FSAL_IS_ERROR(status)) {
-			/* Return success. */
+			/* Return success. We haven't done any permission
+			 * check so ask the caller to do so.
+			 */
+			*caller_perm_check = true;
 			return status;
 		}
 
@@ -1673,10 +1446,10 @@ static fsal_status_t glusterfs_open2(struct fsal_obj_handle *obj_hdl,
 		    ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
 		/* Don't set the mode if we later set the attributes */
-		FSAL_UNSET_MASK(attrib_set->mask, ATTR_MODE);
+		FSAL_UNSET_MASK(attrib_set->valid_mask, ATTR_MODE);
 	}
 
-	if (createmode == FSAL_UNCHECKED && (attrib_set->mask != 0)) {
+	if (createmode == FSAL_UNCHECKED && (attrib_set->valid_mask != 0)) {
 		/* If we have FSAL_UNCHECKED and want to set more attributes
 		 * than the mode, we attempt an O_EXCL create first, if that
 		 * succeeds, then we will be allowed to set the additional
@@ -1710,6 +1483,11 @@ static fsal_status_t glusterfs_open2(struct fsal_obj_handle *obj_hdl,
 		myself = container_of(*new_obj,
 				      struct glusterfs_handle,
 				      handle);
+
+		/* The open is not done with the caller's credentials so ask
+		 * the caller to perform a permission check.
+		 */
+		*caller_perm_check = true;
 		goto open;
 	}
 
@@ -1778,10 +1556,14 @@ static fsal_status_t glusterfs_open2(struct fsal_obj_handle *obj_hdl,
 	 * Also notify caller to do permission check if we DID NOT create the
 	 * file. Note it IS possible in the case of a race between an UNCHECKED
 	 * open and an external unlink, we did create the file, but we will
-	 * still force a permission check. Of course that permission check
-	 * SHOULD succeed since we also won't set the mode the caller requested
-	 * and the default file create permissions SHOULD allow the owner
-	 * read/write.
+	 * still force a permission check. That permission check might fail
+	 * if the file created after the unlink has a mode that doesn't allow
+	 * the caller/creator to open the file (on the other hand, one hopes
+	 * a non-exclusive open doesn't set a mode that doesn't allow read/write
+	 * since the application clearly expects that another process may have
+	 * created the file). This failure case really isn't too awful since
+	 * it would just look to the caller like someone else had created the
+	 * file with a mode that prevented the open this caller was attempting.
 	 */
 	created = (p_flags & O_EXCL) != 0;
 	*caller_perm_check = !created;
@@ -1823,7 +1605,7 @@ open:
 
 	*new_obj = &myself->handle;
 
-	if (created && attrib_set->mask != 0) {
+	if (created && attrib_set->valid_mask != 0) {
 		/* Set attributes using our newly opened file descriptor as the
 		 * share_fd if there are any left to set (mode and truncate
 		 * have already been handled).
@@ -1839,6 +1621,8 @@ open:
 		if (FSAL_IS_ERROR(status)) {
 			/* Release the handle we just allocated. */
 			(*new_obj)->obj_ops.release(*new_obj);
+			/* We released handle at this point */
+			glhandle = NULL;
 			*new_obj = NULL;
 			goto fileerr;
 		}
@@ -1847,7 +1631,7 @@ open:
 			status = (*new_obj)->obj_ops.getattrs(*new_obj,
 							      attrs_out);
 			if (FSAL_IS_ERROR(status) &&
-			    (attrs_out->mask & ATTR_RDATTR_ERR) == 0) {
+			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
 				 * to get the attributes. Otherwise continue
 				 * with attrs_out indicating ATTR_RDATTR_ERR.
@@ -1861,9 +1645,6 @@ open:
 		 * we used to create the fsal_obj_handle.
 		 */
 		posix2fsal_attributes(&sb, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
 	}
 
 
@@ -1920,14 +1701,11 @@ static fsal_status_t glusterfs_reopen2(struct fsal_obj_handle *obj_hdl,
 	fsal_status_t status = {0, 0};
 	int posix_flags = 0;
 	fsal_openflags_t old_openflags;
-	bool truncated;
 
 	my_share_fd = (struct glusterfs_fd *)(state + 1);
 
 	fsal2posix_openflags(openflags, &posix_flags);
 
-	truncated = (posix_flags & O_TRUNC) != 0;
-
 	memset(my_fd, 0, sizeof(*my_fd));
 
 	myself  = container_of(obj_hdl,
@@ -1972,19 +1750,6 @@ static fsal_status_t glusterfs_reopen2(struct fsal_obj_handle *obj_hdl,
 		 */
 		glusterfs_close_my_fd(my_share_fd);
 		*my_share_fd = fd;
-
-		if (truncated) {
-			/* Refresh the attributes */
-			status = glusterfs_fetch_attrs(myself, my_share_fd);
-			if (FSAL_IS_ERROR(status)) {
-				FSAL_CLEAR_MASK(myself->attributes.mask);
-				FSAL_SET_MASK(myself->attributes.mask,
-					       ATTR_RDATTR_ERR);
-				 /** @todo: should handle this
-				   * better.
-				   */
-			}
-		}
 	} else {
 		/* We had a failure on open - we need to revert the share.
 		 * This can block over an I/O operation.
@@ -2180,7 +1945,7 @@ static fsal_status_t glusterfs_commit2(struct fsal_obj_handle *obj_hdl,
 {
 	fsal_status_t status;
 	int retval;
-	struct glusterfs_fd *out_fd = NULL;
+	struct glusterfs_fd tmp_fd = {0, NULL}, *out_fd = &tmp_fd;
 	struct glusterfs_handle *myself = NULL;
 	bool has_lock = false;
 	bool closefd = false;
@@ -2350,6 +2115,7 @@ static fsal_status_t glusterfs_lock_op2(struct fsal_obj_handle *obj_hdl,
 
 	if (retval /* && lock_op == FSAL_OP_LOCK */) {
 		retval = errno;
+		int rc = 0;
 
 		LogDebug(COMPONENT_FSAL,
 			 "fcntl returned %d %s",
@@ -2357,10 +2123,10 @@ static fsal_status_t glusterfs_lock_op2(struct fsal_obj_handle *obj_hdl,
 
 		if (conflicting_lock != NULL) {
 			/* Get the conflicting lock */
-			retval = glfs_posix_lock(my_fd.glfd, F_GETLK,
+			rc = glfs_posix_lock(my_fd.glfd, F_GETLK,
 						 &lock_args);
 
-			if (retval) {
+			if (rc) {
 				retval = errno; /* we lose the initial error */
 				LogCrit(COMPONENT_FSAL,
 					"After failing a lock request, I couldn't even get the details of who owns the lock.");
@@ -2402,72 +2168,11 @@ static fsal_status_t glusterfs_lock_op2(struct fsal_obj_handle *obj_hdl,
 	return fsalstat(posix2fsal_error(retval), retval);
 }
 
-/* getattr2
- */
-
-fsal_status_t glusterfs_getattr2(struct fsal_obj_handle *obj_hdl)
-{
-	struct glusterfs_handle *myself;
-	fsal_status_t status = {0, 0};
-	bool has_lock = false;
-	bool need_fsync = false;
-	bool closefd = false;
-	struct glusterfs_fd my_fd = {0};
-
-	myself = container_of(obj_hdl, struct glusterfs_handle, handle);
-
-#if 0
-	/** @todo: fsid work */
-	if (obj_hdl->fsal != obj_hdl->fs->fsal) {
-		LogDebug(COMPONENT_FSAL,
-			 "FSAL %s operation for handle belonging to FSAL %s, return EXDEV",
-			 obj_hdl->fsal->name, obj_hdl->fs->fsal->name);
-		return fsalstat(posix2fsal_error(EXDEV), EXDEV);
-	}
-#endif
-
-	/* Get a usable file descriptor (don't need to bypass - FSAL_O_ANY
-	 * won't conflict with any share reservation).
-	 */
-	status = find_fd(&my_fd, obj_hdl, false, NULL, FSAL_O_ANY,
-			 &has_lock, &need_fsync, &closefd, false);
-
-	if (FSAL_IS_ERROR(status)) {
-		if (obj_hdl->type == SYMBOLIC_LINK &&
-		    status.major == ERR_FSAL_PERM) {
-			/* You cannot open_by_handle (XFS on linux) a symlink
-			 * and it throws an EPERM error for it.
-			 * open_by_handle_at does not throw that error for
-			 * symlinks so we play a game here.  Since there is
-			 * not much we can do with symlinks anyway,
-			 * say that we did it but don't actually
-			 * do anything.  In this case, return the stat we got
-			 * at lookup time.  If you *really* want to tweek things
-			 * like owners, get a modern linux kernel...
-			 */
-			status = fsalstat(ERR_FSAL_NO_ERROR, 0);
-		}
-		goto out;
-	}
-
-	status = glusterfs_fetch_attrs(myself, &my_fd);
-
- out:
-
-	if (closefd)
-		glusterfs_close_my_fd(&my_fd);
-
-	if (has_lock)
-		PTHREAD_RWLOCK_unlock(&obj_hdl->lock);
-
-	return status;
-}
-
 /**
  * @brief Set attributes on an object
  *
  * This function sets attributes on an object.  Which attributes are
- * set is determined by attrib_set->mask. The FSAL must manage bypass
+ * set is determined by attrib_set->valid_mask. The FSAL must manage bypass
  * or not of share reservations, and a state may be passed.
  *
  * @param[in] obj_hdl    File on which to operate
@@ -2499,7 +2204,7 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 
 	/** @todo: Handle special file symblic links etc */
 	/* apply umask, if mode attribute is to be changed */
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MODE))
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MODE))
 		attrib_set->mode &=
 		    ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
@@ -2518,27 +2223,25 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 	/* Test if size is being set, make sure file is regular and if so,
 	 * require a read/write file descriptor.
 	 */
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_SIZE)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_SIZE)) {
 		if (obj_hdl->type != REGULAR_FILE)
 			return fsalstat(ERR_FSAL_INVAL, EINVAL);
 		openflags = FSAL_O_RDWR;
 	}
 
-	/* Get a usable file descriptor. Share conflict is only possible if
-	 * size is being set. For special files, handle via handle.
-	 */
-	if ((obj_hdl->type == REGULAR_FILE) ||
-		 (obj_hdl->type == DIRECTORY)) {
+	/** TRUNCATE **/
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_SIZE) &&
+	    (obj_hdl->type == REGULAR_FILE)) {
+		/* Get a usable file descriptor. Share conflict is only
+		 * possible if size is being set. For special files,
+		 * handle via handle.
+		 */
 		status = find_fd(&my_fd, obj_hdl, bypass, state, openflags,
 				 &has_lock, &need_fsync, &closefd, false);
 
 		if (FSAL_IS_ERROR(status))
 			goto out;
-	}
 
-	/** TRUNCATE **/
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_SIZE) &&
-	    (obj_hdl->type == REGULAR_FILE)) {
 		retval = glfs_ftruncate(my_fd.glfd, attrib_set->filesize);
 		if (retval != 0) {
 			if (retval != 0) {
@@ -2548,27 +2251,27 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 		}
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MODE)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MODE)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MODE);
 		buffxstat.buffstat.st_mode = fsal2unix_mode(attrib_set->mode);
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_OWNER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_OWNER)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_UID);
 		buffxstat.buffstat.st_uid = attrib_set->owner;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_GROUP)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_GROUP)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_GID);
 		buffxstat.buffstat.st_gid = attrib_set->group;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ATIME)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ATIME)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_ATIME);
 		buffxstat.buffstat.st_atim = attrib_set->atime;
 	}
 
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ATIME_SERVER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ATIME_SERVER)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_ATIME);
 		struct timespec timestamp;
 
@@ -2581,11 +2284,11 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 	}
 
 	/* try to look at glfs_futimens() instead as done in vfs */
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MTIME)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MTIME)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MTIME);
 		buffxstat.buffstat.st_mtim = attrib_set->mtime;
 	}
-	if (FSAL_TEST_MASK(attrib_set->mask, ATTR_MTIME_SERVER)) {
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MTIME_SERVER)) {
 		FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MTIME);
 		struct timespec timestamp;
 
@@ -2601,7 +2304,7 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 	/* EATTRNOTSUPP error.  */
 
 	if (NFSv4_ACL_SUPPORT) {
-		if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ACL)) {
+		if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ACL)) {
 			if (obj_hdl->type == DIRECTORY)
 				buffxstat.is_dir = true;
 			else
@@ -2619,7 +2322,7 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 			/* mode-bits too if not already passed */
 			FSAL_SET_MASK(mask, GLAPI_SET_ATTR_MODE);
 		}
-	} else if (FSAL_TEST_MASK(attrib_set->mask, ATTR_ACL)) {
+	} else if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ACL)) {
 		status = fsalstat(ERR_FSAL_ATTRNOTSUPP, 0);
 		goto out;
 	}
@@ -2897,6 +2600,7 @@ static void handle_to_key(struct fsal_obj_handle *obj_hdl,
 void handle_ops_init(struct fsal_obj_ops *ops)
 {
 	ops->release = handle_release;
+	ops->merge = glusterfs_merge;
 	ops->lookup = lookup;
 	ops->create = create;
 	ops->mkdir = makedir;
@@ -2905,7 +2609,6 @@ void handle_ops_init(struct fsal_obj_ops *ops)
 	ops->symlink = makesymlink;
 	ops->readlink = readsymlink;
 	ops->getattrs = getattrs;
-	ops->setattrs = setattrs;
 	ops->link = linkfile;
 	ops->rename = renamefile;
 	ops->unlink = file_unlink;
